(* Erweiterte Datumsfunktionen
   - Berechnung von Feiertagen
   - zustzliche Konvertierungen und Abfragen

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1  - Jun. 2003 *)

unit DateUtilEx;

interface

uses SysUtils, DateUtils;

type
  THolidayType = (htFix,htEaster,htAdvent,htMonth);
  THolidayView = (hvNone,hvHoliday,hvOther,hvPrivate);

  (* Call-Back-Funktion fr die Ermittlung eines Feiertags *)
  TDateQuery = function (ADate : TDateTime) : THolidayView of object;

  THoliday = record
    Name   : string;
    Active : boolean;
    HView  : THolidayView;
    HType  : THolidayType;
    Day,
    Mult   : smallint;
    Offset : smallint;
    end;

  THolidayRef = record
    Date   : TDateTime;
    Index  : integer;
    end;

const
  defHolidays = 'Holidays.ini';

type
{ ---------------------------------------------------------------- }
(* Verwalten von gesetzliche Feiertagen *)
  THolidayCalendar = class(TObject)
  private
    FYear     : word;
    FDates    : array of THolidayRef;
    FHolidays : array of THoliday;
    FIniName  : string;
    procedure LoadHolidays (Filename : string; New : boolean; DefaultView : THolidayView);
    procedure InitHolidayDates (AYear : word);
    function GetCount : integer;
    procedure SetCount (n : integer);
    function GetHoliday (n : integer) : THoliday;
    procedure SetHoliday (n : integer; Val : THoliday);
  public
    constructor Create (AYear : word);
    constructor CreateFrom (AHolidayCalendar : THolidayCalendar);
    destructor Destroy; override;
    procedure Assign (AHolidayCalendar : THolidayCalendar);
    procedure LoadFromIni (const IniName : string; DefaultView : THolidayView);
    procedure ImportData (const ImpName : string; DefaultView : THolidayView);
    procedure ExportData (const ExpName : string; ViewType : THolidayView);
    procedure ChangeYear (AYear : word);
    procedure Update;
    function GetHolidayView (ADate : TDateTime) : THolidayView;
    function GetHolidayIndex (ADate : TDateTime; First : integer) : integer;
    function GetHolidayNames (ADate : TDateTime) : string;
    property HolidayCount : integer read GetCount write SetCount;
    property Holiday[n : integer] : THoliday read GetHoliday write SetHoliday;
    end;

// Datumstring (dd.mm.[jj]jj) oder dd. mmmm [jj]jj in TDateTime-Wert konvertieren
function ConvertLongDate (DateStr : string) : TDateTime;

// fr DateTime>=0 wie FormatDateTime, sonst Leerstring
function DateToFormatStr (const Format: string; DateTime : TDateTime) : string;

{ ------------------------------------------------------------------- }
// Monat in einem TDateTime-Wert setzen
function SetMonth (Date : TDateTime; Month : integer) : TDateTime;

// Prfen, ob Monat und Jahr bei zwei TDateTime-Werten bereinstimmen
function SameMonth (A,B : TDateTime) : boolean;

// Tag.Datum als String
function DayMonthToStr (ADate : TDateTime) : string;

// Wochentagsnamen: WeekDay 0 Wert nach ISO 8601 ( Montag = 1)
function GetDayName (WeekDay : integer; Long : boolean) : string;

implementation

uses Forms, IniFiles, JrUtils;

{ ---------------------------------------------------------------- }
function DayMonthToStr (ADate : TDateTime) : string;
begin
  Result:=IntToStr(DayOfTheMonth(ADate))+DateSeparator+
          IntToStr(MonthOfTheYear(ADate))+DateSeparator;
  end;

{ ---------------------------------------------------------------- }
// Datumstring (dd.mm.[jj]jj) oder dd. mmmm [jj]jj in TDateTime-Wert konvertieren
function ConvertLongDate (DateStr : string) : TDateTime;
var
  d,m,y : integer;
  s   : string;
  err : boolean;
begin
  Result:=-1;
  err:=false;
  d:=ReadNxtInt(DateStr,DateSeparator,0,err);
  if err then exit;
  DateStr:=TrimLeft(DateStr);
  if Pos(DateSeparator,DateStr)>0 then m:=ReadNxtInt(DateStr,DateSeparator,0,err)
  else begin
    s:=ReadNxtStr(DateStr,' ');
    m:=1;
    while (m<=12) and (s<>LongMonthNames[m]) do inc(m);
    if m>12 then exit;
    end;
  y:=ReadNxtInt(DateStr,' ',0,err);
  if y<100 then if y<50 then y:=y+2000 else y:=y+1900;
  if err then exit;
  Result:=EncodeDate(y,m,d);
  end;

// fr DateTime>=0 wie FormatDateTime, sonst Lerstring
function DateToFormatStr (const Format: string; DateTime : TDateTime) : string;
begin
  if DateTime<0 then Result:=''   // ungltiges Datum
  else Result:=FormatDateTime(Format,DateTime);
  end;

//-----------------------------------------------------------------------------
// Monat in einem TDateTime-Wert setzen
function SetMonth (Date : TDateTime; Month : integer) : TDateTime;
begin
  Result:=IncMonth(Date,Month-MonthOf(Date));
  end;

// Prfen, ob Monat und Jahr bei zwei TDateTime-Werten bereinstimmen
function SameMonth (A,B : TDateTime) : boolean;
begin
  Result:=(MonthOf(A)=MonthOf(B)) and (YearOf(A)=YearOf(B));
  end;

{ ---------------------------------------------------------------- }
// Wochentagsnamen: WeekDay 0 Wert nach ISO 8601 ( Montag = 1)
function GetDayName (WeekDay : integer; Long : boolean) : string;
begin
  WeekDay:=WeekDay mod 7 +1;
  if Long then Result:=LongDayNames[WeekDay]
  else result:=ShortDayNames[WeekDay];
  end;

{ ---------------------------------------------------------------- }
const
  HolSekt = 'Holiday';
  iniCount = 'Count';
  iniName = 'Name';
  iniAct  = 'Active';
  iniView = 'View';
  iniMult = 'Mult';
  iniType = 'Type';
  iniDay  = 'Day';

constructor THolidayCalendar.Create (AYear : word);
begin
  inherited Create;
  (* Feiertage in Deutschland als Vorgabe laden *)
  LoadHolidays (ExtractFilePath(Application.ExeName)+defHolidays,true,hvHoliday);
  InitHolidayDates (AYear);
  FIniName:='';
  end;

constructor THolidayCalendar.CreateFrom (AHolidayCalendar : THolidayCalendar);
begin
  inherited Create;
  Assign(AHolidayCalendar);
  end;

procedure THolidayCalendar.Assign (AHolidayCalendar : THolidayCalendar);
var
  i : integer;
begin
  SetLength(FHolidays,length(AHolidayCalendar.FHolidays));
  for i:=0 to length(AHolidayCalendar.FHolidays)-1 do FHolidays[i]:=AHolidayCalendar.FHolidays[i];
  InitHolidayDates (AHolidayCalendar.FYear);
  FIniName:=AHolidayCalendar.FIniName;
  end;

procedure THolidayCalendar.LoadHolidays (Filename : string; New : boolean; DefaultView : THolidayView);
var
  n,i,j,k : integer;
  s,t     : string;
  ho      : THoliday;
  found   : boolean;
begin
  with TIniFile.Create(Filename) do begin
    n:=ReadInteger(HolSekt+'s',iniCount,0);  // Anzahl Feiertage im Jahr
    if new then j:=0 else j:=length(FHolidays);
    for i:=0 to n-1 do begin  // Feiertage einlesen
      s:=HolSekt+ZStrint(i,3);
      t:=ReadString(s,iniName,'');
      if length(t)>0 then begin
        with ho do begin
          Name:=t;
          Active:=ReadBool(s,iniAct,false);
          HView:=THolidayView(ReadInteger(s,iniView,integer(DefaultView)));
          HType:=THolidayType(ReadInteger(s,iniType,0));
          Mult:=ReadInteger(s,iniMult,0);
          Day:=$101; Offset:=0;
          k:=ReadInteger(s,iniDay,0);
          if (HType=htFix) or (HType=htMonth) then Day:=k else Offset:=k;
          end;
        // bereits vorhanden?
        found:=false;
        for k:=0 to length(FHolidays)-1 do with FHolidays[k] do begin
          if AnsiSameText(Name,ho.Name) then begin
            found:=(HType=ho.HType) and (Mult=ho.Mult);
            if found then begin
              if (HType=htFix) or (HType=htMonth) then found:=Day=ho.Day
              else found:=Offset=ho.Offset;
              end;
            end;
          if found then break;
          end;
        if not found then begin
          SetLength (FHolidays,j+1);
          FHolidays[j]:=ho;
          inc(j);
          end;
        end;
      end;
    Free;
    end;
  end;

destructor THolidayCalendar.Destroy;
var
  n,i     : integer;
  s       : string;
  IniFile : TIniFile;
begin
  if (length(FIniName)>0) then begin
    IniFile:=TIniFile.Create(FIniName);
    with IniFile do begin
      n:=length(FHolidays);
      WriteInteger(HolSekt+'s',iniCount,n);
      for i:=0 to n-1 do with FHolidays[i] do begin  // Feiertage speichern
        s:=HolSekt+ZStrint(i,3);
        EraseSection (s);
        WriteString(s,iniName,Name);
        WriteBool(s,iniAct,Active);
        WriteInteger(s,iniView,integer(HView));
        WriteInteger(s,iniType,integer(HType));
        WriteInteger(s,iniMult,Mult);
        if (HType=htFix) or (HType=htMonth) then WriteInteger(s,iniDay,Day)
        else WriteInteger(s,iniDay,Offset);
        end;
      Free;
      end;
    end;
  FHolidays:=nil;
  FDates:=nil;
  inherited Destroy;
  end;

{------------------------------------------------------------------- }
procedure THolidayCalendar.LoadFromIni(const IniName : string; DefaultView : THolidayView);
begin
  FIniName:=IniName;
  if (length(FIniName)>0) and FileExists(FIniName) then begin
    LoadHolidays (FIniName,true,DefaultView);
    InitHolidayDates (FYear);
    end;
  end;

procedure THolidayCalendar.ImportData (const ImpName : string; DefaultView : THolidayView);
begin
  if FileExists(ImpName) then begin
    LoadHolidays (ImpName,false,DefaultView);
    InitHolidayDates (FYear);
    end;
  end;

procedure THolidayCalendar.ExportData (const ExpName : string; ViewType : THolidayView);
var
  i,j,n : integer;
  s     : string;
begin
  n:=0;
  for i:=0 to length(FHolidays)-1 do if FHolidays[i].HView=ViewType then inc(n);
  if n>0 then with TIniFile.Create(Expname) do begin
    WriteInteger(HolSekt+'s',iniCount,n);
    j:=0;
    for i:=0 to length(FHolidays)-1 do with FHolidays[i] do if HView=ViewType then begin
      s:=HolSekt+ZStrint(j,3);
      inc(j);
      EraseSection(s);
      WriteString(s,iniName,Name);
      WriteBool(s,iniAct,Active);
      WriteInteger(s,iniView,integer(HView));
      WriteInteger(s,iniType,integer(HType));
      WriteInteger(s,iniMult,Mult);
      if (HType=htFix) or (HType=htMonth) then WriteInteger(s,iniDay,Day)
      else WriteInteger(s,iniDay,Offset);
      end;
    Free;
    end;
  end;

{------------------------------------------------------------------- }
// Feiertagsdaten berechnen fr AYear und AYear+1
procedure THolidayCalendar.InitHolidayDates (AYear : word);
var
  i,j,k,n : word;
  dt   : TDateTime;
  x,y : integer;

  (* Osterdatum berechnen - Formel nach PTB *)
  procedure OsterDatum (jahr : word; var monat,tag : word);
  var
    k,m,a,d,og,sz : integer;
  begin
    k:=jahr div 100;
    m:=15+(3*k+3) div 4 -(8*k+13) div 25;
    a:=jahr mod 19;
    d:=(19*a+m) mod 30;
    (* og = Mrzdatum des Ostervollmonds (= 14. Tag des ersten
       Monats im Mondkalender, genannt Nisanu)*)
    og:=21+d-(d div 29)-(d div 28 - d div 29)*(a div 11);
    (* sz = Datum des 1. Sonntags im Mrz *)
    sz:=7-(jahr+(jahr div 4)+2-(3*k+3) div 4) mod 7;
    tag:=og+7-((og-sz) mod 7); monat:=3;
    if tag>31 then begin
      tag:=tag-31; monat:=4;
      end;
    end;

  function GetEasterDate (Year : word) : TDateTime;
  var
    m,t : word;
  begin
    OsterDatum(Year,m,t);
    Result:=EncodeDate(FYear+i,m,t);
    end;

begin
  FYear:=AYear;
  n:=length(FHolidays);
  SetLength(FDates,2*n);
  for i:=0 to 1 do begin
    j:=AYear+i;
    if n>0 then for k:=0 to n-1 do with FHolidays[k] do begin
      case HType of
      htFix: with FDates[i*n+k] do begin
            Date:=EncodeDate(j,(Day shr 8)and $FF,Day and $FF);
            Index:=k;
            end;
      htEaster: with FDates[i*n+k] do begin
            Date:=GetEasterDate(j)+Offset;
            Index:=k;
            end;
      htAdvent : with FDates[i*n+k] do begin
            dt:=EncodeDate(j,12,24);
            Date:=dt-DayOfWeek(dt)-20+Offset;
            Index:=k;
            end;
      htMonth : with FDates[i*n+k] do begin
            case Day shr 4 and $3 of
            1 : begin // letzter Wochentag im Monat
                dt:=EncodeDate(j,(Day shr 8)and $FF,1);
                dt:=dt+DaysInMonth(dt)-1;
                x:=(Day and $F); y:=DayOfTheWeek(dt)-1;
                Date:=dt-(7+y-x) mod 7;
                end;
            2 : begin // 2. Wochentag im Monat
                dt:=EncodeDate(j,(Day shr 8)and $FF,1);
                x:=(Day and $F); y:=DayOfTheWeek(dt)-1;
                Date:=dt+7+(7+x-y) mod 7;
                end;
            3 : begin // vorletzter Wochentag im Monat
                dt:=EncodeDate(j,(Day shr 8)and $FF,1);
                dt:=dt+DaysInMonth(dt)-1;
                x:=(Day and $F); y:=DayOfTheWeek(dt)-1;
                Date:=dt-7-(7+y-x) mod 7;
                end;
              else begin // 1. Wochentag im Monat
                dt:=EncodeDate(j,(Day shr 8)and $FF,1);
                x:=(Day and $F); y:=DayOfTheWeek(dt)-1;
                Date:=dt+(7+x-y) mod 7;
                end;
              end;
            Index:=k;
            end;
        end;
      end;
    end;
  end;

{------------------------------------------------------------------- }
function THolidayCalendar.GetCount : integer;
begin
  Result:=length(FHolidays);
  end;

procedure THolidayCalendar.SetCount (n : integer);
begin
  SetLength(FHolidays,n);
  end;

function THolidayCalendar.GetHoliday (n : integer) : THoliday;
begin
  if (n>=0) and (n<length(FHolidays)) then Result:=FHolidays[n]
  else Result.Name:='';
  end;

procedure THolidayCalendar.SetHoliday (n : integer; Val : THoliday);
begin
  if (n>=0) and (n<length(FHolidays)) then FHolidays[n]:=Val;
  end;

{------------------------------------------------------------------- }
procedure THolidayCalendar.Update;
begin
  InitHolidayDates (FYear);
  end;

procedure THolidayCalendar.ChangeYear (AYear : word);
begin
  if AYear<>FYear then InitHolidayDates (AYear);
  end;

function THolidayCalendar.GetHolidayView (ADate : TDateTime) : THolidayView;
var
  year : word;
  i    : integer;
  h    : THolidayView;
begin
  year:=YearOf(ADate);
  if (year<FYear) or (year>FYear+1) then InitHolidayDates (year);
  h:=hvNone;
  for i:=0 to length(FDates)-1 do with FDates[i] do with FHolidays[Index] do if Active then begin
    if Mult<>0 then begin
      if Mult<0 then begin
        if (trunc(Date+Mult-1)<trunc(ADate)) and (trunc(ADate)<=trunc(Date)) then h:=HView;
        end
      else if (trunc(Date)<=trunc(ADate)) and (trunc(ADate)<trunc(Date+Mult+1)) then h:=HView;
      end
    else if SameDate(Date,ADate) then begin
      Result:=HView; exit;
      end;
    end;
  Result:=h;
  end;

function THolidayCalendar.GetHolidayIndex (ADate : TDateTime; First : integer) : integer;
var
  i,n  : integer;
begin
  if YearOf(ADate)=FYear then begin
    n:=-1;
    for i:=First to length(FDates)-1 do with FDates[i] do with FHolidays[Index] do if Active then begin
      if Mult<>0 then begin
        if Mult<0 then begin
          if (trunc(Date+Mult-1)<trunc(ADate)) and (trunc(ADate)<=trunc(Date)) then n:=i;
          end
        else if (trunc(Date)<=trunc(ADate)) and (trunc(ADate)<trunc(Date+Mult+1)) then n:=i;
        end
      else if SameDate(Date,ADate) then begin
        Result:=i; exit;
        end;
      end;
    Result:=n;
    end
  else Result:=-1;
  end;

function THolidayCalendar.GetHolidayNames (ADate : TDateTime) : string;
var
  i : integer;
begin
  Result:='';
  if YearOf(ADate)=FYear then begin
    for i:=0 to length(FDates)-1 do with FDates[i] do with FHolidays[Index] do if Active then begin
      if Mult<>0 then begin
        if Mult<0 then begin
          if (trunc(Date+Mult-1)<trunc(ADate)) and (trunc(ADate)<=trunc(Date)) then Result:=Result+','+Name;
          end
        else if (trunc(Date)<=trunc(ADate)) and (trunc(ADate)<trunc(Date+Mult+1)) then Result:=Result+','+Name;
        end
      else if SameDate(Date,ADate) then Result:=Result+','+Name;
      end;
    Delete(Result,1,1);
    end;
  end;

end.
