(* Global types, constants and subroutines for
   ExtFileTools, AesLib, FtpFileTools

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1.0 - Jun. 2006
   vers. 2.0 - Apr. 2007 : Error messages changed
   *)

unit FtGlobal;

interface

type
  TFileAction = (acNone,acCopy,acCompress,acUnCompress,acEncrypt,acDecrypt,
                 acVerify,acAddVerify,acFtpWrite,acFtpRead);
  TProgressEvent = procedure(AAction : TFileAction; ACount: int64) of object;

const
  // error codes
  errOK         = 0;
  errFileCreate = 1;
  errFileOpen   = 2;
  errFileClose  = 3;
  errFileRead   = 4;
  errFileWrite  = 5;
  errFileAttr   = 6;  // Error setting attr.
  errFileFull   = 7;
  errFileGZip   = 8;  // ill. GZip header
  errFileCheck  = 9;  // corrupt copied or packed file
  errFileECrypt = 10; // error encrypting file
  errFileDCrypt = 11; // error decrypting file
  errFileVerify = 12; // verify error
  errLongPath   = 13; // path too long (>260)
  errNotFound   = 14; // source file not found
  errFileTS     = 15; // Error setting timestamp after copy
  errStorage    = 16; // Error copying file details
  errFtpRead    = 24; // error reading via FTP
  errFtpWrite   = 25; // error writing via FTP
  errFtpConnect = 26; // connection error on FTP
  errFtpBroken  = 27; // connection reset by server
  errFTPDatConn = 28; // error opening data connection
  errSignature  = 32; // ill. Zip signature
  errExtract    = 33; // Error extracting file
  errFormat     = 34; // Unsupported file format

  errUserBreak  = $80; // process stopped bei user
  errAllCodes   = $00FF;

  errError      = $0100;
  errWarning    = $0200;
  errAllTypes   = $0F00;

  errCopy       = $1000;
  errGZip       = $2000;
  errGUnzip     = $3000;
  errZip        = $4000;
  errEncrypt    = $5000;
  errDecrypt    = $6000;
  errAllSources = $F000;
  
function rsErrOpening : string;
function rsErrCreating : string;
function rsErrReading : string;
function rsErrWriting : string;
function rsErrClosing : string;
function rsErrSetAttr : string;
function rsErrTimeStamp : string;
function rsErrDirEntry : string;
function rsErrEndRecord : string;
function rsErrVerify : string;

// get error message
function GetCopyErrMsg (AError : integer) : string;

implementation

uses SysUtils;

{ ------------------------------------------------------------------- }
// get error message
function rsErrOpening : string;
begin
  Result:='Fehler beim ffnen von %s';
  end;

function rsErrCreating : string;
begin
  Result:='Fehler beim Erstellen von %s';
  end;

function rsErrReading : string;
begin
  Result:='Fehler beim Lesen aus %s';
  end;

function rsErrWriting : string;
begin
  Result:='Fehler beim Schreiben in %s';
  end;

function rsErrClosing : string;
begin
  Result:='Fehler beim Schlieen von %s';
  end;

function rsErrSetAttr : string;
begin
  Result:='Fehler beim Setzen der Attribute von %s';
  end;

function rsErrTimeStamp : string;begin
  Result:='Fehler beim Setzen des Zeitstempels von %s';
  end;

function rsErrDirEntry : string;
begin
  Result:='Fehler beim Erstellen eines Verzeichniseintrags in %s';
  end;

function rsErrEndRecord : string;
begin
  Result:='Fehler beim Schreiben des Endblocks in %s';
  end;

function rsErrVerify : string;
begin
  Result:='Fehler beim berprfen von %s';
  end;

function GetCopyErrMsg (AError : integer) : string;
var
  s : string;
begin
  case AError and errAllTypes of
  errError   : s:='Fehler: ';
  errWarning : s:='Warnung: ';
  else s:='Info: '
    end;
  case AError and errAllCodes of
  errFileCreate : s:=s+'Datei konnte nicht erstellt werden';
  errFileOpen   : s:=s+'Datei konnte nicht geffnet werden';
  errFileClose  : s:=s+'Datei konnte nicht geschlossen werden';
  errFileRead   : s:=s+'Aus der Datei konnte nicht gelesen werden';
  errFileWrite  : s:=s+'In die Datei konnte nicht geschrieben werden';
  errFileAttr   : s:=s+'Die Dateiattribute konnten nicht gesetzt werden';
  errFileFull   : s:=s+'Zu wenig Platz auf dem Datentrger';
  errFileGZip   : s:=s+'Ungltiger GZip Header';
  errFileCheck  : s:=s+'Beschdigte Datei';
  errFileECrypt : s:=s+'Die Verschlsselung ist fehlgeschlagen';
  errFileDCrypt : s:=s+'Die Entschlsselung ist fehlgeschlagen';
  errFileVerify : s:=s+'Die berprfung ist fehlgeschlagen';
  errLongPath   : s:=s+'Pfad zu lang';
  errNotFound   : s:=s+'Datei nicht gefunden';
  errFileTS     : s:=s+'Der Zeitstempel konnte nicht gesetzt werden';
  errFtpRead    : s:=s+'ber FTP konnte nicht gelesen werden';
  errFtpWrite   : s:=s+'ber FTP konnte nicht geschrieben werden';
  errFtpConnect : s:=s+'Die FTP-Verbindung konnte nicht aufgebaut werden';
  errFtpBroken  : s:=s+'Die FTP-Verbindung wurde vom Server beendet';
  errFtpDatConn : s:=s+'Fehler beim Herstellen der FTP-Datenverbindung';
  errSignature  : s:=s+'Ungltige Zip-Signatur';
  errExtract    : s:=s+'Fehler beim Extrahieren einer Datei';
  errFormat     : s:=s+'Nicht untersttztes Dateiformat';
  errUserBreak  : s:=s+'Abbruch durch den Benutzer';
  else s:=s+'Unbekannter Fehler ('+IntToHex(AError,8)+')';  // should not happen
    end;
  case AError and errAllSources of
  errCopy       : s:=s+' (Kopieren)';
  errGZip       : s:=s+' (GZ packen)';
  errGUnzip     : s:=s+' (GZ entpacken)';
  errZip        : s:=s+' (Zip packen)';
  errEnCrypt    : s:=s+' (Verschlsseln)';
  errDeCrypt    : s:=s+' (Entschlsseln)';
    end;
  Result:=s;
  end;

end.
 
